@@ -44,7 +44,7 @@ impl Connecting {
         conn_events: mpsc::UnboundedReceiver<ConnectionEvent>,
         udp_state: Arc<UdpState>,
         runtime: Arc<dyn Runtime>,
-    ) -> Connecting {
+    ) -> Self {
         let (on_handshake_data_send, on_handshake_data_recv) = oneshot::channel();
         let (on_connected_send, on_connected_recv) = oneshot::channel();
         let conn = ConnectionRef::new(
@@ -60,7 +60,7 @@ impl Connecting {
 
         runtime.spawn(Box::pin(ConnectionDriver(conn.clone())));
 
-        Connecting {
+        Self {
             conn: Some(conn),
             connected: on_connected_recv,
             handshake_data_ready: Some(on_handshake_data_recv),
@@ -157,6 +157,14 @@ impl Connecting {
 
         inner.inner.local_ip()
     }
+
+    /// The peer's UDP address.
+    ///
+    /// Will panic if called after `poll` has returned `Ready`.
+    pub fn remote_address(&self) -> SocketAddr {
+        let conn_ref: &ConnectionRef = self.conn.as_ref().expect("used after yielding Ready");
+        conn_ref.state.lock("remote_address").inner.remote_address()
+    }
 }
 
 impl Future for Connecting {
@@ -178,16 +186,6 @@ impl Future for Connecting {
     }
 }
 
-impl Connecting {
-    /// The peer's UDP address.
-    ///
-    /// Will panic if called after `poll` has returned `Ready`.
-    pub fn remote_address(&self) -> SocketAddr {
-        let conn_ref: &ConnectionRef = self.conn.as_ref().expect("used after yielding Ready");
-        conn_ref.state.lock("remote_address").inner.remote_address()
-    }
-}
-
 /// Future that completes when a connection is fully established
 ///
 /// For clients, the resulting value indicates if 0-RTT was accepted. For servers, the resulting
@@ -738,7 +736,7 @@ impl Future for ReadDatagram<'_> {
 }
 
 #[derive(Debug)]
-pub struct ConnectionRef(Arc<ConnectionInner>);
+pub(crate) struct ConnectionRef(Arc<ConnectionInner>);
 
 impl ConnectionRef {
     #[allow(clippy::too_many_arguments)]
@@ -813,7 +811,7 @@ impl std::ops::Deref for ConnectionRef {
 }
 
 #[derive(Debug)]
-pub struct ConnectionInner {
+pub(crate) struct ConnectionInner {
     pub(crate) state: Mutex<State>,
     pub(crate) shared: Shared,
 }
@@ -1080,7 +1078,7 @@ impl State {
     }
 
     /// Close for a reason other than the application's explicit request
-    pub fn implicit_close(&mut self, shared: &Shared) {
+    pub(crate) fn implicit_close(&mut self, shared: &Shared) {
         self.close(0u32.into(), Bytes::new(), shared);
     }
 
@@ -1149,6 +1147,6 @@ pub struct UnknownStream {
 
 impl From<proto::UnknownStream> for UnknownStream {
     fn from(_: proto::UnknownStream) -> Self {
-        UnknownStream { _private: () }
+        Self { _private: () }
     }
 }
