@@ -88,10 +88,10 @@ impl Endpoint {
         config: EndpointConfig,
         server_config: Option<ServerConfig>,
         socket: std::net::UdpSocket,
-        runtime: impl Runtime,
+        runtime: Arc<dyn Runtime>,
     ) -> io::Result<Self> {
         let socket = runtime.wrap_udp_socket(socket)?;
-        Self::new_with_runtime(config, server_config, socket, Arc::new(runtime))
+        Self::new_with_runtime(config, server_config, socket, runtime)
     }
 
     /// Construct an endpoint with arbitrary configuration and pre-constructed abstract socket
@@ -102,9 +102,9 @@ impl Endpoint {
         config: EndpointConfig,
         server_config: Option<ServerConfig>,
         socket: impl AsyncUdpSocket,
-        runtime: impl Runtime,
+        runtime: Arc<dyn Runtime>,
     ) -> io::Result<Self> {
-        Self::new_with_runtime(config, server_config, Box::new(socket), Arc::new(runtime))
+        Self::new_with_runtime(config, server_config, Box::new(socket), runtime)
     }
 
     fn new_with_runtime(
@@ -114,9 +114,10 @@ impl Endpoint {
         runtime: Arc<dyn Runtime>,
     ) -> io::Result<Self> {
         let addr = socket.local_addr()?;
+        let allow_mtud = !socket.may_fragment();
         let rc = EndpointRef::new(
             socket,
-            proto::Endpoint::new(Arc::new(config), server_config.map(Arc::new)),
+            proto::Endpoint::new(Arc::new(config), server_config.map(Arc::new), allow_mtud),
             addr.is_ipv6(),
             runtime.clone(),
         );
@@ -235,6 +236,21 @@ impl Endpoint {
         self.inner.state.lock().unwrap().socket.local_addr()
     }
 
+    /// Reject new incoming connections without affecting existing connections
+    ///
+    /// Convenience short-hand for using
+    /// [`set_server_config`](Self::set_server_config) to update
+    /// [`concurrent_connections`](ServerConfig::concurrent_connections) to
+    /// zero.
+    pub fn reject_new_connections(&self) {
+        self.inner
+            .state
+            .lock()
+            .unwrap()
+            .inner
+            .reject_new_connections();
+    }
+
     /// Close all of this endpoint's connections immediately and cease accepting new connections.
     ///
     /// See [`Connection::close()`] for details.
@@ -271,8 +287,10 @@ impl Endpoint {
                 if endpoint.connections.is_empty() {
                     break;
                 }
+                // Construct future while lock is held to avoid race
+                self.inner.shared.idle.notified()
             }
-            self.inner.shared.idle.notified().await;
+            .await;
         }
     }
 }
@@ -348,7 +366,7 @@ pub(crate) struct State {
     socket: Box<dyn AsyncUdpSocket>,
     udp_state: Arc<UdpState>,
     inner: proto::Endpoint,
-    outgoing: VecDeque<proto::Transmit>,
+    outgoing: VecDeque<udp::Transmit>,
     incoming: VecDeque<Connecting>,
     driver: Option<Waker>,
     ipv6: bool,
@@ -361,6 +379,8 @@ pub(crate) struct State {
     recv_buf: Box<[u8]>,
     send_limiter: WorkLimiter,
     runtime: Arc<dyn Runtime>,
+    /// The packet contents length in the outgoing queue.
+    outgoing_queue_contents_len: usize,
 }
 
 #[derive(Debug)]
@@ -392,10 +412,13 @@ impl State {
                         let mut data: BytesMut = buf[0..meta.len].into();
                         while !data.is_empty() {
                             let buf = data.split_to(meta.stride.min(data.len()));
-                            match self
-                                .inner
-                                .handle(now, meta.addr, meta.dst_ip, meta.ecn, buf)
-                            {
+                            match self.inner.handle(
+                                now,
+                                meta.addr,
+                                meta.dst_ip,
+                                meta.ecn.map(proto_ecn),
+                                buf,
+                            ) {
                                 Some((handle, DatagramEvent::NewConnection(conn))) => {
                                     let conn = self.connections.insert(
                                         handle,
@@ -447,7 +470,7 @@ impl State {
         let result = loop {
             while self.outgoing.len() < BATCH_SIZE {
                 match self.inner.poll_transmit() {
-                    Some(x) => self.outgoing.push_back(x),
+                    Some(t) => self.queue_transmit(t),
                     None => break,
                 }
             }
@@ -465,7 +488,9 @@ impl State {
                 .poll_send(&self.udp_state, cx, self.outgoing.as_slices().0)
             {
                 Poll::Ready(Ok(n)) => {
-                    self.outgoing.drain(..n);
+                    let contents_len: usize =
+                        self.outgoing.drain(..n).map(|t| t.contents.len()).sum();
+                    self.decrement_outgoing_contents_len(contents_len);
                     // We count transmits instead of `poll_send` calls since the cost
                     // of a `sendmmsg` still linearily increases with number of packets.
                     self.send_limiter.record_work(n);
@@ -506,7 +531,7 @@ impl State {
                                 .send(ConnectionEvent::Proto(event));
                         }
                     }
-                    Transmit(t) => self.outgoing.push_back(t),
+                    Transmit(t) => self.queue_transmit(t),
                 },
                 Poll::Ready(None) => unreachable!("EndpointInner owns one sender"),
                 Poll::Pending => {
@@ -517,6 +542,52 @@ impl State {
 
         true
     }
+
+    fn queue_transmit(&mut self, t: proto::Transmit) {
+        let contents_len = t.contents.len();
+        self.increment_outgoing_queue_contents_len(contents_len);
+        self.outgoing.push_back(udp::Transmit {
+            destination: t.destination,
+            ecn: t.ecn.map(udp_ecn),
+            contents: t.contents,
+            segment_size: t.segment_size,
+            src_ip: t.src_ip,
+        });
+    }
+
+    fn increment_outgoing_queue_contents_len(&mut self, contents_len: usize) {
+        self.outgoing_queue_contents_len = self
+            .outgoing_queue_contents_len
+            .saturating_add(contents_len);
+        self.inner
+            .set_socket_buffer_fill(self.outgoing_queue_contents_len);
+    }
+
+    fn decrement_outgoing_contents_len(&mut self, contents_len: usize) {
+        self.outgoing_queue_contents_len = self
+            .outgoing_queue_contents_len
+            .saturating_sub(contents_len);
+        self.inner
+            .set_socket_buffer_fill(self.outgoing_queue_contents_len);
+    }
+}
+
+#[inline]
+fn udp_ecn(ecn: proto::EcnCodepoint) -> udp::EcnCodepoint {
+    match ecn {
+        proto::EcnCodepoint::Ect0 => udp::EcnCodepoint::Ect0,
+        proto::EcnCodepoint::Ect1 => udp::EcnCodepoint::Ect1,
+        proto::EcnCodepoint::Ce => udp::EcnCodepoint::Ce,
+    }
+}
+
+#[inline]
+fn proto_ecn(ecn: udp::EcnCodepoint) -> proto::EcnCodepoint {
+    match ecn {
+        udp::EcnCodepoint::Ect0 => proto::EcnCodepoint::Ect0,
+        udp::EcnCodepoint::Ect1 => proto::EcnCodepoint::Ect1,
+        udp::EcnCodepoint::Ce => proto::EcnCodepoint::Ce,
+    }
 }
 
 #[derive(Debug)]
@@ -640,6 +711,7 @@ impl EndpointRef {
                 recv_limiter: WorkLimiter::new(RECV_TIME_BOUND),
                 send_limiter: WorkLimiter::new(SEND_TIME_BOUND),
                 runtime,
+                outgoing_queue_contents_len: 0,
             }),
         }))
     }
